////////////////////////////////////////////////////////////////////////
//  rpud : An R package for GPU computing
//  Copyright (C) 2010-2014 Chi Yau. All Rights Reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


#include "rpud.h"
#include "inc/rpudev.h"


enum RPUDIST_TYPE {
    RPUDIST_UNKNOWN,
    RPUDIST_EUCLIDEAN,
    RPUDIST_MAXIMUM,
    RPUDIST_MANHATTAN,
    RPUDIST_CANBERRA,
    RPUDIST_BINARY,
    RPUDIST_MINKOWSKI
};


////////////////////////////////////////////////////////////////////////
// rpuMetricKernel
//
inline __device__ size_t distoffset(
    const size_t n, const size_t x, const size_t y) {

    return y+x*n-((x+1)*(x+2))/2;
}


namespace rpudist {

// 2D texture
static texture<Float, 2, cudaReadModeElementType> 
    textRef;     

template <typename T, RPUDIST_TYPE t>
__global__ void rpuMetricKernel(
    T* d_output, size_t num, size_t dim, T p) {
        
    const int 
        xid = blockIdx.x * blockDim.x + threadIdx.x,
        yid = blockIdx.y * blockDim.y + threadIdx.y;
        
    // compute the distance between the x-th and y-th vectors
    if (xid < yid && yid < num) {

        size_t num_item = 0;
        T dist = 0;
        for (size_t i = 0; i < dim; i++) {
        
            const T 
                x = tex2D(textRef, xid, i),
                y = tex2D(textRef, yid, i);
            if (RPU_BOTH_NON_NA(x, y)) {
                switch (t) {
                
                case RPUDIST_EUCLIDEAN:
                {
                const T dev = x - y;
                if (RPU_NON_NA(dev)) {
                    dist += dev * dev;
                    num_item++;
                } }
                break;
                
                case RPUDIST_MAXIMUM:
                {
                const T dev = abs(x - y);
                if (RPU_NON_NA(dev)) {
                    if (dist < dev) dist = dev;
                    num_item++;
                } }
                break;
                
                case RPUDIST_MANHATTAN:
                {
                const T dev = abs(x - y);
                if (RPU_NON_NA(dev)) {
                    dist += dev;
                    num_item++;
                } }
                break;
                
                case RPUDIST_CANBERRA:
                {
                const T 
                    sum = abs(x + y),   // [Bug 14375]
                    diff = abs(x - y);
                if (sum > RPU_NUMERIC_MIN_DENORM || 
                    diff > RPU_NUMERIC_MIN_DENORM) {
                    T dev = diff/sum;
                    if (RPU_NON_NA(dev)) {
                        dist += dev;
                        num_item++;
                    } else if (!RPU_FINITE(diff) && diff == sum) {
                        dist += 1;
                        num_item++;
                    }
                } }
                break;
                
                case RPUDIST_BINARY:
                {
                if (RPU_BOTH_FINITE(x, y)) {
                    if (x || y) {
                        num_item++;
                        if (!(x && y)) dist++;
                    }
                } }
                break;
                
                case RPUDIST_MINKOWSKI:
                {
                const T dev = (x - y);
                if (RPU_NON_NA(dev)) {
                    dist += __powf(abs(dev), p);
                    num_item++;
                } }
                break;
                
                } // switch t
                
            } // if RPU

            
        } // for i
         
        if (num_item != dim) {
            dist *= T(dim)/T(num_item);
        }

        d_output[distoffset(num, xid, yid)] = 
            (t == RPUDIST_EUCLIDEAN) ? sqrt(dist) :
            (t == RPUDIST_MINKOWSKI) ? __powf(dist, T(1.0/p)) :
            dist;
            
    } // if xid

}


////////////////////////////////////////////////////////////////////////
// RPUDist
//
template <typename T>
class RPUDist {

public:
    RPUDist(const char* type, const T* input, 
        const size_t num, const size_t dim, const T power);
     
    void find(T* output);

protected:
    const T* 
        m_input;
    const size_t
        m_nx, m_ny, m_num,
        BLK_DX, BLK_DY;          // block dimensions
    const T
        m_power;

    RPUDIST_TYPE
        m_type;
};


template <typename T>
RPUDist<T>::RPUDist(const char* type, const T* input, 
        const size_t num, const size_t dim, const T power) :
            m_input(input), m_nx(num), m_ny(dim), m_num(num), 
            m_power(power), BLK_DX(16), BLK_DY(16) {

    m_type =
        (!strcmp("euclidean", type))    ? RPUDIST_EUCLIDEAN :
        (!strcmp("maximum", type))      ? RPUDIST_MAXIMUM   :
        (!strcmp("manhattan", type))    ? RPUDIST_MANHATTAN :
        (!strcmp("canberra", type))     ? RPUDIST_CANBERRA  :
        (!strcmp("binary", type))       ? RPUDIST_BINARY    :
        (!strcmp("minkowski", type))    ? RPUDIST_MINKOWSKI :
        RPUDIST_UNKNOWN;
        
    if (m_type == RPUDIST_UNKNOWN) {
        throw rpu::AppException("Undefined distance type");
    }

}


template <typename T>
void RPUDist<T>::find(T* h_output) {

    RPU_MSG("RPUDist::find starts\n");
     
    const char *szError = NULL;
    cudaArray* cuArray = NULL;
    PFloat d_output = NULL;
     
    try {
        ///////////////////////////////////////////////////////////////////
        // device input
        const size_t INPUT_SIZE = sizeof(T)*(m_nx)*(m_ny);
        rpuSafeCall(cudaMallocArray(&cuArray, &textRef.channelDesc, m_nx, m_ny));

        // copy input from host to device
        rpuSafeCall(cudaMemcpyToArray(cuArray, 0, 0, m_input, INPUT_SIZE, cudaMemcpyHostToDevice));

        // bind texture to array
        rpuSafeCall(cudaBindTextureToArray(textRef, cuArray));

        ///////////////////////////////////////////////////////////////////
        // device output
        const size_t OUTPUT_SIZE = sizeof(T)*(m_num)*(m_num-1)/2;
        rpuSafeCall(cudaMalloc((void**)&d_output, OUTPUT_SIZE));
    
        ///////////////////////////////////////////////////////////////////
        // kernel call
        const dim3 
            grid((m_num+BLK_DX-1)/BLK_DX, (m_num+BLK_DY-1)/BLK_DY), 
            blk(BLK_DX, BLK_DY);
                
        switch (m_type) {
        case RPUDIST_EUCLIDEAN:
            rpuMetricKernel<T, RPUDIST_EUCLIDEAN><<< grid, blk >>>(
                d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_MAXIMUM:
            rpuMetricKernel<T, RPUDIST_MAXIMUM><<< grid, blk >>>(
                d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_MANHATTAN:
            rpuMetricKernel<T, RPUDIST_MANHATTAN><<< grid, blk >>>(
                d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_CANBERRA:
            rpuMetricKernel<T, RPUDIST_CANBERRA><<< grid, blk >>>(
                d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_BINARY:
            rpuMetricKernel<T, RPUDIST_BINARY><<< grid, blk >>>(
                d_output, m_nx, m_ny, m_power
                );
            break;
    
        case RPUDIST_MINKOWSKI:
            rpuMetricKernel<T, RPUDIST_MINKOWSKI><<< grid, blk >>>(
                d_output, m_nx, m_ny, m_power
                );
            break;

        default:
            assert(false);
            break;
        }
        rpuCheckMsg("Kernel execution failed");

        // block until the device has completed
        rpuSafeCall(cudaDeviceSynchronize());

        // copy device output to host
        rpuSafeCall(cudaMemcpy(h_output, d_output, OUTPUT_SIZE, cudaMemcpyDeviceToHost));
    
        RPU_MSG("RPUDist::find ends\n");
     
    } catch (const std::exception& e) {
        szError = e.what();
    } catch (...) {
        szError = "Unknown error";
    }

    ///////////////////////////////////////////////////////////////////
    // free device memory
    cudaUnbindTexture(textRef);
    if (cuArray != NULL) cudaFreeArray(cuArray);
    if (d_output != NULL) cudaFree(d_output);
    
    ///////////////////////////////////////////////////////////////////
    // exit
    if (szError != NULL) error(szError);
    
}


}   // namespace rpudist


////////////////////////////////////////////////////////////////////////
// rpuDistance
//
void rpuDistance(
        const PString pType, 
        const PFloat points, 
        const PInteger pNum, 
        const PInteger pDim,
        const PFloat pMinkowski,
        PFloat pDistances) {

    String type = *pType;
    RPU_MSG("Finding %s distances ...\n", type);
     
    rpudist::RPUDist<Float> dist(
        type, 
        points, 
        static_cast<size_t>(*pNum), 
        static_cast<size_t>(*pDim), 
        *pMinkowski
        );
    dist.find(pDistances);
}






